# bzr-dbus: dbus support for bzr/bzrlib.
# Copyright (C) 2007 Canonical Limited.
#   Author: Robert Collins.
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
# 

"""Activity of bzr.

This module provides Activity which supports announcing and recieving messages
about bzr activity: See the class for more detail.
"""

import socket
import time

import dbus.bus
import dbus.lowlevel
import gobject

from bzrlib import errors
from bzrlib.plugins.dbus import mapper
from bzrlib.revision import NULL_REVISION
from bzrlib.smart.protocol import _encode_tuple, _decode_tuple


# Dont try to put a '-' in bazaar-vcs here, its NOT dns and dbus
# considers - illegal.
BROADCAST_INTERFACE = "org.bazaarvcs.plugins.dbus.Broadcast"


def _get_bus(bus):
    """Get a bus thats usable."""
    if bus is None:
        # lazy import to not pollute bzr during startup.
        import dbus.mainloop.glib
        dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
        return dbus.SessionBus()
    else:
        return bus


def send_signal(bus, dbus_interface, signal_name, signature, *args):
    """Send a signal on the bus."""
    message = dbus.lowlevel.SignalMessage('/', dbus_interface, signal_name)
    message.append(signature=signature, *args)
    bus.send_message(message)


class Activity(object):
    """bzrlib object activity tracking."""

    def __init__(self, bus=None):
        """Create an Activity object.

        :param bus: A dbus Bus object. By default this will be set to
            bus.SessionBus(). If you need a private bus or a system bus,
            supply this parameter.
        """
        self.bus = _get_bus(bus)

    def advertise_branch(self, branch):
        """Advertise branch to dbus.

        This is a top level convenience function to advertise a branch. No
        warranties are made about delivery of the advertisement, nor of how
        long it will be visible to users. Specifically, dbus errors are caught,
        and the advertisement is not repeated.
        :param branch: The branch to be advertised. The advertisement is done
            by announcing the tip revision and the URL of the branch.
        :return: None
        :raises: Nothing should be raised.
        """
        self.announce_revision_urls(branch.last_revision(), [branch.base])

    def announce_revision_urls(self, revision, urls):
        """Low level revision-specific announce logic. 

        The recommended API is advertise_branch, announce_revision, while 
        public is not stable or supported. Use at your own warranty.

        This method does not translate urls: its expected that the urls
        being advertised are already translated.
        """
        if revision in (None, NULL_REVISION):
            revision = '' # avoid sending None or NULL_REVISION (its internal
                          # only) on the wire.
        send_signal(self.bus, BROADCAST_INTERFACE, 'Revision',
                    'sas', revision, urls)


class LanGateway(object):
    """A gateway for bazaar commit notifications to the local lan."""
    DBUS_NAME = "org.bazaarvcs.plugins.dbus.LanGateway"

    def __init__(self, bus=None, mainloop=None, activity=None):
        """Create a LanGateway object.

        :param bus: A dbus Bus object. By default this will be set to
            bus.SessionBus(). If you need a private bus or a system bus,
            supply this parameter.
        """
        self.bus = _get_bus(bus)
        if self.bus.request_name(
            self.DBUS_NAME, dbus.bus.NAME_FLAG_DO_NOT_QUEUE) not in [
            dbus.bus.REQUEST_NAME_REPLY_ALREADY_OWNER,
            dbus.bus.REQUEST_NAME_REPLY_PRIMARY_OWNER]:
            raise errors.BzrError(
                "A lan-notify process is already running")
        if mainloop is None:
            self.mainloop = gobject.MainLoop()
        else:
            self.mainloop = mainloop
        if activity is None:
            self.activity = Activity(self.bus)
        else:
            self.activity = activity
        self.url_mapper = mapper.URLMapper()
        self.seen_revisions = {}

    def broadcast_data(self, data):
        """Transmit data on the LAN in a broadcast packet."""
        self.sock.sendto(data, ('<broadcast>', 4155))

    def catch_dbus_revision(self, revision_id, urls, sender_name):
        """Catch a published revision_id from dbus."""
        # Ignore revisions emitted by us to prevent loops.
        if sender_name == self.bus.get_unique_name():
            return
        packet_args = ['announce_revision', revision_id]
        seen_urls = set()
        for revisions in self.seen_revisions.values():
            if revision_id in revisions:
                seen_urls.update(set(revisions[revision_id]))
        allowed_urls = []
        # Pass URLs through the mapper
        for unmapped_url in urls:
            mapped_urls = self.url_mapper.map(unmapped_url)
            if not mapped_urls:
                mapped_urls = [unmapped_url]
            for url in mapped_urls:
                if not url.startswith('file:///') and not url in seen_urls:
                    allowed_urls.append(url)
        self.note_network_revision(time.time(), revision_id, allowed_urls)
        packet_args.extend(allowed_urls)
        if len(packet_args) == 2:
            # no valid urls.
            return
        self.broadcast_data(_encode_tuple(packet_args))

    def handle_network_data(self, data):
        """Handle data from the network."""
        args = _decode_tuple(data)
        assert args[0] == 'announce_revision'
        rev_id = args[1]
        urls = args[2:]
        announce = True
        for revisions in self.seen_revisions.values():
            if rev_id in revisions:
                announce = False
        self.note_network_revision(time.time(), rev_id, urls)
        if announce:
            self.activity.announce_revision_urls(rev_id, urls)

    def handle_network_packet(self, source, condition):
        """Handle a packet from the network."""
        data = source.recvfrom(65535)
        self.handle_network_data(data[0])
        return True

    def note_network_revision(self, now, revid, urls):
        """Note that revid was seen at urls now.

        This will remove stale cache entries.

        :param now: the result of time.time().
        """
        keys_to_remove = []
        for key in self.seen_revisions.keys():
            if key + 5 < int(now/60):
                keys_to_remove.append(key)
        for key in keys_to_remove:
            del self.seen_revisions[key]
        self.seen_revisions.setdefault(int(now/60), {})[revid] = urls

    def run(self, _port=4155):
        """Start the LanGateway.
        
        The optional _port parameter is used for testing.
        """
        self.start(_port)
        try:
            self.mainloop.run()
        finally:
            self.stop()

    def server_started(self, backing_urls, public_url):
        """Add a url prefix to be mapped when advertising revisions."""
        for backing_url in backing_urls:
            self.url_mapper.add_map(backing_url, public_url)

    def server_stopped(self, backing_urls, public_url):
        """Remove a previously mapped url prefix."""
        for backing_url in backing_urls:
            self.url_mapper.remove_map(backing_url, public_url)

    def start(self, _port=4155):
        """Start the LanGateway.
        
        The optional _port parameter is used for testing.
        """
        # listen for network events
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        self.sock.bind(('', _port))
        if _port == 0:
            self.port = self.sock.getsockname()[1]
        else:
            self.port = _port
        self.watch_id = gobject.io_add_watch(
            self.sock, gobject.IO_IN, self.handle_network_packet)
        # Add listeners for ServerStarted and ServerStopped signals,
        # then ask all servers to re-advertise themselves.
        self.bus.add_signal_receiver(
            self.server_started,
            dbus_interface=BROADCAST_INTERFACE,
            signal_name='ServerStarted')
        self.bus.add_signal_receiver(
            self.server_stopped,
            dbus_interface=BROADCAST_INTERFACE,
            signal_name='ServerStopped')
        send_signal(self.bus, BROADCAST_INTERFACE, 'QueryServers', '')
        # listen for revisions on the bus
        self.bus.add_signal_receiver(
            self.catch_dbus_revision,
            dbus_interface=BROADCAST_INTERFACE,
            signal_name='Revision',
            sender_keyword='sender_name')

    def stop(self):
        """Stop the LanGateway."""
        gobject.source_remove(self.watch_id)
        self.sock.close()
        self.sock = None


class ServerAnnouncer(object):

    def __init__(self, bus=None):
        self.bus = _get_bus(bus)
        self.servers = set()
        self.bus.add_signal_receiver(
            self.query_servers,
            dbus_interface=BROADCAST_INTERFACE,
            signal_name='QueryServers')

    def emit(self, signal_name, backing_urls, public_url):
        send_signal(self.bus, BROADCAST_INTERFACE, signal_name,
                    'ass', backing_urls, public_url)

    def on_server_started(self, backing_urls, public_url):
        self.servers.add((tuple(backing_urls), public_url))
        self.emit('ServerStarted', backing_urls, public_url)

    def on_server_stopped(self, backing_urls, public_url):
        self.emit('ServerStopped', backing_urls, public_url)
        try:
            self.servers.remove((tuple(backing_urls), public_url))
        except KeyError:
            pass

    def query_servers(self):
        for backing_urls, public_url in self.servers:
            self.emit('ServerStarted', backing_urls, public_url)
