# bzr-dbus: dbus support for bzr/bzrlib.
# Copyright (C) 2007 Canonical Limited.
#   Author: Robert Collins.
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
# 

"""Tests for the dbus activity service."""

import os
import signal
import socket
import subprocess
import tempfile
import thread
import time
import weakref

import dbus
import dbus.bus
import dbus.mainloop.glib
import gobject

from bzrlib import errors
import bzrlib.plugins
from bzrlib.smart.protocol import _encode_tuple, _decode_tuple

# done here to just do it once, and not in the plugin module to avoid doing it
# by default.
dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)

from bzrlib.tests import TestCaseWithMemoryTransport

from bzrlib.plugins.dbus import activity


def create_daemon():
    """Create a dbus daemon."""
    config_file = tempfile.NamedTemporaryFile()
    config_file.write('''
<!DOCTYPE busconfig PUBLIC "-//freedesktop//DTD D-Bus Bus Configuration 1.0//EN"
 "http://www.freedesktop.org/standards/dbus/1.0/busconfig.dtd">
<busconfig>
  <!-- Our well-known bus type, don't change this -->
  <type>session</type>

  <listen>unix:tmpdir=/tmp</listen>

  <policy context="default">
    <!-- Allow everything to be sent -->
    <allow send_destination="*"/>
    <!-- Allow everything to be received -->
    <allow eavesdrop="true"/>
    <!-- Allow anyone to own anything -->
    <allow own="*"/>
  </policy>

</busconfig>
''')
    config_file.flush()
    proc = subprocess.Popen(
        ['dbus-daemon', '--config-file=%s' % config_file.name,
        '--print-address'], stdout=subprocess.PIPE)
    address = proc.stdout.readline()
    return address.strip(), proc


class TemporaryBus(dbus.bus.BusConnection):
    """A TemporaryBus created within this application."""

    def __new__(cls):
        address, proc = create_daemon()
        try:
            result = dbus.bus.BusConnection.__new__(cls, address)
        except:
            os.kill(proc.pid, signal.SIGINT)
            raise
        result._test_process = proc
        result._bus_type = address
        return result

    def nuke(self):
        """Ensure the daemon is shutdown."""
        os.kill(self._test_process.pid, signal.SIGINT)


class TestCaseWithDBus(TestCaseWithMemoryTransport):

    def setUp(self):
        TestCaseWithMemoryTransport.setUp(self)
        # setup a private dbus session so we dont spam 
        # the users desktop!
        self.bus = TemporaryBus()
        self.addCleanup(self.bus.nuke)


class TestActivity(TestCaseWithDBus):
        
    def test_advertise_branch_no_service_running(self):
        # should not error: just call it
        activity.Activity(bus=self.bus).advertise_branch(self.make_branch('.'))

    def test_advertise_branch_service_running_no_commits(self):
        """No commits in a branch leads to it being show with a revid of ''."""
        # we want to recieve the callbacks to inspect them.
        branches = []
        def revision_cb(revision_id, urls):
            branches.append((revision_id, urls))
            mainloop.quit()
        self.bus.add_signal_receiver(
            revision_cb,
            dbus_interface=activity.BROADCAST_INTERFACE,
            signal_name='Revision')

        # now call our convenience JustDoIt method.
        branch = self.make_branch('.')
        activity.Activity(bus=self.bus).advertise_branch(branch)
        # now, let the broadcast method interactions all happen

        # '' because the branch has no commits.
        mainloop = gobject.MainLoop()
        mainloop.run()
        self.assertEqual([('', [branch.base])], branches)

    def test_advertise_branch_service_running_with_commits(self):
        """With commits, the commit is utf8 encoded."""
        # we want to recieve the callbacks to inspect them.
        branches = []
        def revision_cb(revision_id, urls):
            branches.append((revision_id, urls))
            mainloop.quit()
        self.bus.add_signal_receiver(
            revision_cb,
            dbus_interface=activity.BROADCAST_INTERFACE,
            signal_name='Revision')

        # now call our convenience JustDoIt method.
        tree = self.make_branch_and_memory_tree('.')
        tree.lock_write()
        tree.add('')
        tree.commit('commit', rev_id=u'\xc8'.encode('utf8'))
        tree.unlock()
        activity.Activity(bus=self.bus).advertise_branch(tree.branch)
        # now, let the broadcast method interactions all happen

        # in theory, the revid here would be utf8 encoded, but dbus seems to
        # consider 'string' == 'unicode', and 'Magic happens' as far as wire
        # level decoding -> unicode results. CRY.
        mainloop = gobject.MainLoop()
        mainloop.run()
        self.assertEqual([(u'\xc8', [tree.branch.base])], branches)

    def test_constructor(self):
        """The constructor should setup a good working environment."""
        obj = activity.Activity()
        # dbus uses a singleton approach for SessionBus, and we dont know if
        # it has __eq__ or if it would be reliable, so we use object identity.
        self.assertTrue(obj.bus is dbus.SessionBus(),
            "%r is not %r" % (obj.bus, dbus.SessionBus()))


class TestLanGateway(TestCaseWithDBus):

    # catch urls from network
    # urls from network: add to time-limited-cache, put on dbus.
    # deserialise content

    def test_create(self):
        gateway = activity.LanGateway(self.bus)
        # if no mainloop was supplied, a new one is created.
        self.assertNotEqual(None, gateway.mainloop)

    def test_only_one_gateway(self):
        """Only one LanGateway may be active for a bus"""
        gateway1 = activity.LanGateway(self.bus)

        # Create a second connection to the bus
        second_bus_connection = dbus.bus.BusConnection(self.bus._bus_type)
        try:
            self.assertRaises(errors.BzrError, activity.LanGateway,
                              second_bus_connection)
        finally:
            second_bus_connection.close()

    def test_run_binds(self):
        self.error = None
        mainloop = gobject.MainLoop()
        # we want run to: open a listening socket on localhost:4155 udp, the
        # bzr protocol port number.
        # to test this we want to check the port is in use during the loop.
        def check_socket_bound():
            # try to bind to the socket which should be already bound.
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            try:
                try:
                    self.assertRaises(socket.error, sock.bind, ('127.0.0.1', gateway.port))
                except Exception, e:
                    self.error = e
            finally:
                gobject.timeout_add(0, mainloop.quit)
        gobject.timeout_add(0, check_socket_bound)
        gateway = activity.LanGateway(self.bus, mainloop)
        # disable talking to dbus by making it a noop
        gateway.activity.listen_for_revisions = lambda x:x
        gateway.run(_port=0)
        if self.error:
            raise self.error

    def test_network_packets_trigger_handle_network_packet(self):
        mainloop = gobject.MainLoop()
        # we want network packets to call handle_network_packet,
        # so we override that to shutdown the mainloop and log the call.
        def send_packet():
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
            sock.sendto('data_to_handle', ('127.0.0.1', gateway.port))
            sock.close()
        gobject.timeout_add(0, send_packet)
        gobject.timeout_add(1000, mainloop.quit)
        gateway = activity.LanGateway(self.bus, mainloop)
        calls = []
        def handle_data(data):
            calls.append(('handle', data))
            gobject.timeout_add(0, mainloop.quit)
        gateway.handle_network_data = handle_data
        # disable talking to dbus by making it a noop
        gateway.activity.listen_for_revisions = lambda x:x
        gateway.run(_port=0)
        self.assertEqual([('handle', 'data_to_handle')], calls)

    def test_broadcast_data_calls_sock(self):
        # the method broadcast_data on a LanGateway should invoke
        # socket.sendto(data, ('<broadcast>', 4155))
        mainloop = gobject.MainLoop()
        gateway = activity.LanGateway(self.bus, mainloop)
        calls = []
        class StubSocket(object):
            def sendto(self, data, address):
                calls.append(('sendto', data, address))
        gateway.sock = StubSocket()
        gateway.broadcast_data('some data')
        self.assertEqual([('sendto', 'some data', ('<broadcast>', 4155))],
            calls)

    def test_run_listens_for_revisions(self):
        calls = []
        class StubBus(object):
            def request_name(self, name, flags):
                return dbus.bus.REQUEST_NAME_REPLY_PRIMARY_OWNER
            def send_message(self, message):
                pass
            def add_signal_receiver(self, callback, dbus_interface,
                                    signal_name, sender_keyword=None):
                calls.append((callback, dbus_interface, signal_name,
                              sender_keyword))
        mainloop = gobject.MainLoop()
        gobject.timeout_add(0, mainloop.quit)
        gateway = activity.LanGateway(StubBus(), mainloop)
        gateway.run(0)
        self.assertEqual(calls[-1], (gateway.catch_dbus_revision,
                                     activity.BROADCAST_INTERFACE,
                                     'Revision', 'sender_name'))

    def test_catch_dbus_revision_ignore_file_only(self):
        """catch_dbus_revision should ignore file:/// urls."""
        mainloop = gobject.MainLoop()
        gateway = activity.LanGateway(self.bus, mainloop)
        # instrument the transmission apis used by the LanGateway.
        calls = []
        def broadcast_data(data):
            calls.append(('broadcast_data', data))
        gateway.broadcast_data = broadcast_data
        gateway.catch_dbus_revision('a revid', ['file:///'], None)
        self.assertEqual([], calls)

    def test_catch_dbus_revision_strip_file(self):
        """catch_dbus_revision should strip file:/// urls if others exist."""
        mainloop = gobject.MainLoop()
        gateway = activity.LanGateway(self.bus, mainloop)
        # instrument the transmission apis used by the LanGateway.
        calls = []
        def broadcast_data(data):
            calls.append(('broadcast_data', data))
        gateway.broadcast_data = broadcast_data
        gateway.catch_dbus_revision(
            'revid', ['file:///foo/', 'http://bar'], None)
        self.assertEqual(
            [('broadcast_data',
              _encode_tuple(('announce_revision', 'revid', 'http://bar')))],
            calls)

    def test_catch_dbus_revision_encodes_smart_tuple(self):
        """catch_dbus_revision should encode as _encode_tuple does."""
        mainloop = gobject.MainLoop()
        gateway = activity.LanGateway(self.bus, mainloop)
        # instrument the transmission apis used by the LanGateway.
        calls = []
        def broadcast_data(data):
            calls.append(('broadcast_data', data))
        gateway.broadcast_data = broadcast_data
        gateway.catch_dbus_revision(
            'a revid', ['http://bar', 'bzr://host/'], None)
        self.assertEqual(
            [('broadcast_data',
              _encode_tuple(('announce_revision', 'a revid', 'http://bar', 'bzr://host/')))],
            calls)

    def test_catch_dbus_revision_ignores_revision_url_all_from_network(self):
        """There is a time limited window within which revisions are remembered."""
        mainloop = gobject.MainLoop()
        gateway = activity.LanGateway(self.bus, mainloop)
        # instrument the transmission apis used by the LanGateway.
        calls = []
        def broadcast_data(data):
            calls.append(('broadcast_data', data))
        gateway.broadcast_data = broadcast_data
        now = time.time()
        gateway.note_network_revision(now, 'a revid', ['url1', 'url2'])
        gateway.catch_dbus_revision('a revid', ['url1', 'url2'], None)
        self.assertEqual([], calls)

    def test_catch_dbus_revision_preserves_non_network_urls(self):
        """When some urls for a revision were not from the network the rest are sent."""
        mainloop = gobject.MainLoop()
        gateway = activity.LanGateway(self.bus, mainloop)
        # instrument the transmission apis used by the LanGateway.
        calls = []
        def broadcast_data(data):
            calls.append(('broadcast_data', data))
        gateway.broadcast_data = broadcast_data
        now = time.time()
        gateway.note_network_revision(now, 'a revid', ['url1'])
        gateway.catch_dbus_revision('a revid', ['url2'], None)
        self.assertEqual(
            [('broadcast_data',
              _encode_tuple(('announce_revision', 'a revid', 'url2')))],
            calls)

    def test_catch_dbus_revision_notes_revision(self):
        """Revisions coming in from dbus are noted too, to prevent reemission."""
        mainloop = gobject.MainLoop()
        gateway = activity.LanGateway(self.bus, mainloop)
        # instrument the transmission apis used by the LanGateway.
        calls = []
        def broadcast_data(data):
            calls.append(('broadcast_data', data))
        def note_network_revision(now, revid, urls):
            calls.append(('note', now, revid, urls))
        gateway.broadcast_data = broadcast_data
        gateway.note_network_revision = note_network_revision
        start = time.time()
        gateway.catch_dbus_revision('a revid', ['url2'], None)
        finish = time.time()
        self.assertEqual(
            [('note', calls[0][1], 'a revid', ['url2']),
             ('broadcast_data',
              _encode_tuple(('announce_revision', 'a revid', 'url2'))),
             ],
            calls)
        self.assertTrue(start <= calls[0][1])
        self.assertTrue(calls[0][1] <= finish)

    def test_note_network_revision(self):
        mainloop = gobject.MainLoop()
        gateway = activity.LanGateway(self.bus, mainloop)
        now = time.time()
        # noting the we've seen 'rev' now should cache it at minute granularity.
        gateway.note_network_revision(now, 'rev', ['url1', 'url2'])
        self.assertEqual({int(now/60):{'rev':['url1', 'url2']}},
            gateway.seen_revisions)

    def test_note_network_revision_trims_cache(self):
        mainloop = gobject.MainLoop()
        gateway = activity.LanGateway(self.bus, mainloop)
        now = time.time()
        # noting a time when there are stale entries should remove them.
        gateway.note_network_revision(now - 60*6, 'rev', ['url1', 'url2'])
        gateway.note_network_revision(now - 60*4, 'rev1', ['url3', 'url4'])
        gateway.note_network_revision(now, 'rev2', ['url5', 'url6'])
        self.assertEqual({
            int((now-60*4)/60):{'rev1':['url3', 'url4']},
            int(now/60):{'rev2':['url5', 'url6']},
            },
            gateway.seen_revisions)

    def test_handle_network_data(self):
        """data from the network is deserialised, passed to note_revision and dbus."""

        mainloop = gobject.MainLoop()
        # instrument the transmission apis used by the LanGateway.
        calls = []
        start = time.time()
        def note_network_revision(now, revid, urls):
            calls.append(('note_rev', now, revid, urls))
        class StubActivity(object):
            def announce_revision_urls(self, revision, urls):
                calls.append(('announce', revision, urls))
        an_activity = StubActivity()
        gateway = activity.LanGateway(self.bus, mainloop, an_activity)
        gateway.note_network_revision = note_network_revision
        network_data = _encode_tuple(('announce_revision', 'a revid', 'url2'))
        gateway.handle_network_data(network_data)
        finish = time.time()
        self.assertEqual(
            [('note_rev', calls[0][1], 'a revid', ('url2',)),
             ('announce', 'a revid', ('url2',))],
             calls)
        self.assertTrue(start <= calls[0][1])
        self.assertTrue(calls[0][1] <= finish)

    def test_emits_query_servers(self):
        """The broadcaster emits a QueryServers signal on startup."""
        # Add a signal listener for the QueryServers signal.
        calls = []
        def query_servers_cb():
            calls.append(None)
            mainloop.quit()
        self.bus.add_signal_receiver(
            query_servers_cb,
            dbus_interface=activity.BROADCAST_INTERFACE,
            signal_name='QueryServers')
        gateway = activity.LanGateway(self.bus)
        gateway.start(0)
        mainloop = gobject.MainLoop()
        mainloop.run()
        self.assertEqual([None], calls)
        gateway.stop()

    def test_handles_server_started(self):
        """The LanGateway listens for ServerStarted signals."""
        calls = []
        class MyLanGateway(activity.LanGateway):
            def server_started(self, backing_urls, public_url):
                super(MyLanGateway, self).server_started(
                    backing_urls, public_url)
                calls.append((backing_urls, public_url))
                if len(calls) == 2:
                    mainloop.quit()

        # attach a Broadcaster.
        gateway = MyLanGateway(self.bus)
        gateway.start(0)
        # emit the signals.
        activity.send_signal(self.bus, activity.BROADCAST_INTERFACE,
                             'ServerStarted', 'ass', ['foo/'], 'bar/')
        # for two locations at the same url
        activity.send_signal(self.bus, activity.BROADCAST_INTERFACE,
                             'ServerStarted', 'ass', ['foo/'], 'baz/')
        mainloop = gobject.MainLoop()
        mainloop.run()
        self.assertEqual([(['foo/'], 'bar/'), (['foo/'], 'baz/')], calls)
        self.assertEqual({'foo/':['bar/', 'baz/']}, gateway.url_mapper.maps)
        gateway.stop()

    def test_handles_server_stopped(self):
        """The LanGateway listens for ServerStopped signals."""
        calls = []
        class MyLanGateway(activity.LanGateway):
            def server_started(self, backing_urls, public_url):
                super(MyLanGateway, self).server_started(
                    backing_urls, public_url)
                calls.append(('ServerStarted', backing_urls, public_url))
                if len(calls) == 4:
                    mainloop.quit()

            def server_stopped(self, backing_urls, public_url):
                super(MyLanGateway, self).server_stopped(
                    backing_urls, public_url)
                calls.append(('ServerStopped', backing_urls, public_url))
                if len(calls) == 4:
                    mainloop.quit()

        # attach a Broadcaster.
        gateway = MyLanGateway(self.bus)
        gateway.start(0)

        # add two locations at the same prefix
        activity.send_signal(self.bus, activity.BROADCAST_INTERFACE,
                             'ServerStarted', 'ass', ['foo/'], 'bar/')
        activity.send_signal(self.bus, activity.BROADCAST_INTERFACE,
                             'ServerStarted', 'ass', ['foo/'], 'baz/')
        # and remove them both
        activity.send_signal(self.bus, activity.BROADCAST_INTERFACE,
                             'ServerStopped', 'ass', ['foo/'], 'bar/')
        activity.send_signal(self.bus, activity.BROADCAST_INTERFACE,
                             'ServerStopped', 'ass', ['foo/'], 'baz/')

        mainloop = gobject.MainLoop()
        mainloop.run()
        self.assertEqual([('ServerStarted', ['foo/'], 'bar/'),
                          ('ServerStarted', ['foo/'], 'baz/'),
                          ('ServerStopped', ['foo/'], 'bar/'),
                          ('ServerStopped', ['foo/'], 'baz/')], calls)
        self.assertEqual({}, gateway.url_mapper.maps)
        gateway.stop()


# TODO: garbage from the network should be handled cleanly.


class ServerAnnouncerTests(TestCaseWithDBus):

    def test_create(self):
        announcer = activity.ServerAnnouncer(self.bus)
        self.assertEqual(self.bus, announcer.bus)
        self.assertEqual(set(), announcer.servers)

    def test_on_server_started(self):
        """on_server_started() sends a DBus signal."""
        mainloop = gobject.MainLoop()
        calls = []
        def server_started(backing_urls, public_url):
            calls.append((backing_urls, public_url))
            mainloop.quit()
        self.bus.add_signal_receiver(
            server_started,
            dbus_interface=activity.BROADCAST_INTERFACE,
            signal_name='ServerStarted')
        announcer = activity.ServerAnnouncer(self.bus)
        announcer.on_server_started(['backing-url'], 'public-url')
        # Run the mainloop to wait for signals.
        mainloop.run()
        self.assertEqual(1, len(calls))
        self.assertEqual((['backing-url'], 'public-url'), calls[0])
        self.assertTrue((('backing-url',), 'public-url') in announcer.servers)

    def test_on_server_stopped(self):
        """on_server_stopped() sends a DBus signal."""
        mainloop = gobject.MainLoop()
        calls = []
        def server_stopped(backing_urls, public_url):
            calls.append((backing_urls, public_url))
            mainloop.quit()
        self.bus.add_signal_receiver(
            server_stopped,
            dbus_interface=activity.BROADCAST_INTERFACE,
            signal_name='ServerStopped')
        announcer = activity.ServerAnnouncer(self.bus)
        announcer.servers.add((('backing-url',), 'public-url'))
        announcer.on_server_stopped(['backing-url'], 'public-url')
        # Run the mainloop to wait for signals.
        mainloop.run()
        self.assertEqual(1, len(calls))
        self.assertEqual((['backing-url'], 'public-url'), calls[0])
        self.assertTrue((('backing-url',), 'public-url')
                        not in announcer.servers)

    def test_query_servers(self):
        """All servers are reannounced in response to QueryServers."""
        mainloop = gobject.MainLoop()
        calls = []
        def server_started(backing_urls, public_url):
            calls.append((backing_urls, public_url))
            if len(calls) == 2:
                mainloop.quit()
        self.bus.add_signal_receiver(
            server_started,
            dbus_interface=activity.BROADCAST_INTERFACE,
            signal_name='ServerStarted')
        announcer = activity.ServerAnnouncer(self.bus)
        # Add some server entries
        announcer.servers.add((('backing-url-1',), 'public-url-1'))
        announcer.servers.add((('backing-url-2', 'second-url'), 'public-url-2'))
        activity.send_signal(
            self.bus, activity.BROADCAST_INTERFACE,
            'QueryServers', '')
        # Run the mainloop, which will quit when 2 ServerStarted
        # signals are received.
        mainloop.run()
        self.assertEqual(2, len(calls))
        calls.sort()
        self.assertEqual((['backing-url-1'], 'public-url-1'), calls[0])
        self.assertEqual((['backing-url-2', 'second-url'], 'public-url-2'),
                         calls[1])
