# bzr-dbus: dbus support for bzr/bzrlib.
# Copyright (C) 2007 Canonical Limited.
#   Author: Robert Collins.
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
# 

"""Tests for the dbus hooks."""

from bzrlib.branch import Branch
from bzrlib.smart.server import SmartTCPServer

from bzrlib.plugins.dbus import activity, hook, server_mainloop
from bzrlib.plugins.dbus.tests.test_activity import TestCaseWithDBus


class TestHooksAreSet(TestCaseWithDBus):

    def setUp(self):
        super(TestHooksAreSet, self).setUp()
        self._mainloop_hooks = server_mainloop.MainloopThread.hooks
        server_mainloop.MainloopThread.hooks = \
            server_mainloop.MainloopThreadHooks()
        hook._announcer = None

    def tearDown(self):
        server_mainloop.MainloopThread.hooks = self._mainloop_hooks
        super(TestHooksAreSet, self).tearDown()

    def test_hooks_are_installed(self):
        """Loading the plugin should have installed its hooks."""
        # check by looking in self._preserved hooks.
        # the set_rh Branch hook to detect branch changes.
        self.assertTrue(hook.on_post_change_branch_tip in
            self._preserved_hooks[Branch]['post_change_branch_tip'])
        # the server_started and server_stopped smart server hooks
        # to detect url maps for servers.
        self.assertTrue(hook.on_server_start in
                        self._mainloop_hooks['server_started'])
        self.assertTrue(hook.on_server_stop in
                        self._mainloop_hooks['server_stopped'])

    def test_install_hooks(self):
        """dbus.hook.install_hooks() should install hooks."""
        hook.install_hooks()
        # check the branch hooks.
        self.assertTrue(hook.on_post_change_branch_tip in
                        Branch.hooks['post_change_branch_tip'])
        # check the SmartServer hooks.
        self.assertTrue(hook.on_server_start in
                        server_mainloop.MainloopThread.hooks['server_started'])
        self.assertTrue(hook.on_server_stop in
                        server_mainloop.MainloopThread.hooks['server_stopped'])

    def test_on_post_change_branch_tip_hook(self):
        """The on_post_change_branch_tip hook should advertise the branch."""
        calls = []
        class SampleActivity(object):

            def advertise_branch(self, branch):
                calls.append(('advertise_branch', branch))

        class FakeParams(object):
            branch = 'branch'

        # prevent api skew: check we can use the API SampleActivity presents.
        activity.Activity(bus=self.bus).advertise_branch(self.make_branch('.'))
        # now test the hook
        original_class = activity.Activity
        try:
            activity.Activity = SampleActivity
            hook.on_post_change_branch_tip(FakeParams)
        finally:
            activity.Activity = original_class
        self.assertEqual([('advertise_branch', 'branch')], calls)

    def test_on_server_start_hook(self):
        """The on_server_start hook should add a URL mapping for the server."""
        # change the global b.p.dbus.activity.ServerAnnouncer to instrument
        # on_server_start.
        calls = []
        class SampleServerAnnouncer(object):

            def on_server_started(self, backing_urls, public_url):
                calls.append(('on_server_started', backing_urls, public_url))

        # prevent api skew: check we can use the API
        # SampleServerAnnouncer presents.
        activity.ServerAnnouncer(bus=self.bus).on_server_started(
            ['backing-url'], 'public-url')
        # now test the hook
        original_class = activity.ServerAnnouncer
        try:
            activity.ServerAnnouncer = SampleServerAnnouncer
            hook.on_server_start(['source'], 'target')
        finally:
            activity.ServerAnnouncer = original_class
        self.assertEqual([('on_server_started', ['source'], 'target')], calls)

    def test_on_server_stop_hook(self):
        """The on_server_stop hook should add a URL mapping for the server."""
        # change the global b.p.dbus.activity.ServerAnnouncer to instrument
        # on_server_stop.
        calls = []
        class SampleServerAnnouncer(object):

            def on_server_stopped(self, backing_urls, public_url):
                calls.append(('on_server_stopped', backing_urls, public_url))

        # prevent api skew: check we can use the API
        # SampleServerAnnouncer presents.
        activity.ServerAnnouncer(bus=self.bus).on_server_stopped(
            ['backing-url'], 'public-url')
        # now test the hook
        original_class = activity.ServerAnnouncer
        try:
            activity.ServerAnnouncer = SampleServerAnnouncer
            hook.on_server_stop(['source'], 'target')
        finally:
            activity.ServerAnnouncer = original_class
        self.assertEqual([('on_server_stopped', ['source'], 'target')], calls)
