import threading

from bzrlib.plugins.dbus import server_mainloop
from bzrlib.smart.server import SmartTCPServer
from bzrlib.tests import TestCase


class MainloopThreadTests(TestCase):

    def setUp(self):
        super(MainloopThreadTests, self).setUp()
        self._mainloop_hooks = server_mainloop.MainloopThread.hooks
        server_mainloop.MainloopThread.hooks = \
            server_mainloop.MainloopThreadHooks()
        server_mainloop._mainloop_thread = None

    def tearDown(self):
        server_mainloop.MainloopThread.hooks = self._mainloop_hooks
        super(MainloopThreadTests, self).tearDown()

    def test_start_stop(self):
        """The start() and stop() methods control the mainloop thread."""
        mainloop = server_mainloop.MainloopThread()
        self.assertEqual(None, mainloop.thread)
        mainloop.start()
        self.assertNotEqual(None, mainloop.thread)
        mainloop.stop()
        self.assertEqual(None, mainloop.thread)

    def test_recursive_start_stop(self):
        """The start() and stop() methods are recursive."""
        mainloop = server_mainloop.MainloopThread()
        self.assertEqual(0, mainloop.start_count)
        mainloop.start()
        thread = mainloop.thread
        self.assertEqual(1, mainloop.start_count)
        mainloop.start()
        # Still the same thread:
        self.assertEqual(thread, mainloop.thread)
        self.assertEqual(2, mainloop.start_count)
        # Must call stop() twice to kill the thread.
        mainloop.stop()
        self.assertEqual(1, mainloop.start_count)
        mainloop.stop()
        self.assertEqual(0, mainloop.start_count)
        self.assertEqual(None, mainloop.thread)

    def test_thread_start_stop_hooks(self):
        """Hooks are called when the mainloop thread starts and stops."""
        # Install some hooks:
        calls = []
        def on_thread_started(mainloop):
            calls.append(('thread_started', mainloop,
                          threading.currentThread()))
        def on_thread_stopped(mainloop):
            calls.append(('thread_stopped', mainloop,
                          threading.currentThread()))
        server_mainloop.MainloopThread.hooks.install_hook(
            'thread_started', on_thread_started)
        server_mainloop.MainloopThread.hooks.install_hook(
            'thread_stopped', on_thread_stopped)
        mainloop = server_mainloop.MainloopThread()
        mainloop.start()
        # The thread_started hook has been called in the other thread.
        self.assertEqual(1, len(calls))
        self.assertEqual(('thread_started', mainloop), calls[0][:2])
        self.assertNotEqual(threading.currentThread(), calls[0][2])
        # Calling start() again does not trigger the hook.
        mainloop.start()
        self.assertEqual(1, len(calls))
        # The thread_stopped hook is called when the thread exits
        mainloop.stop()
        self.assertEqual(1, len(calls))
        mainloop.stop()
        self.assertEqual(2, len(calls))
        self.assertEqual(('thread_stopped', mainloop), calls[1][:2])
        self.assertNotEqual(threading.currentThread(), calls[1][2])
        self.assertEqual(0, mainloop.start_count)
        self.assertEqual(None, mainloop.thread)

    def test_server_start_stop_hooks(self):
        """The server_started and server_stopped hooks are mirrored by
        MainloopThread.
        """
        calls = []
        def on_thread_started(mainloop):
            calls.append(('thread_started', mainloop,
                          threading.currentThread()))
        def on_thread_stopped(mainloop):
            calls.append(('thread_stopped', mainloop,
                          threading.currentThread()))
        def on_server_started(backing_urls, public_url):
            calls.append(('server_started', backing_urls,
                          public_url, threading.currentThread()))
        def on_server_stopped(backing_urls, public_url):
            calls.append(('server_stopped', backing_urls,
                          public_url, threading.currentThread()))
        server_mainloop.MainloopThread.hooks.install_hook(
            'thread_started', on_thread_started)
        server_mainloop.MainloopThread.hooks.install_hook(
            'thread_stopped', on_thread_stopped)
        server_mainloop.MainloopThread.hooks.install_hook(
            'server_started', on_server_started)
        server_mainloop.MainloopThread.hooks.install_hook(
            'server_stopped', on_server_stopped)

        # on_start_server() starts the mainloop and calls a similar
        # hook in the thread.
        server_mainloop.on_server_started(['backing-url-1'], 'public-url-1')
        self.assertEqual(2, len(calls))
        self.assertEqual('thread_started', calls[0][0])
        self.assertTrue(isinstance(calls[0][1],
                                   server_mainloop.MainloopThread))
        self.assertNotEqual(threading.currentThread(), calls[0][2])
        mainloop = calls[0][1]
        thread = calls[0][2]
        self.assertEqual(('server_started', ['backing-url-1'],
                          'public-url-1', thread), calls[1])

        # Starting another server runs the hook again in the same
        # thread with the same mainloop
        server_mainloop.on_server_started(['backing-url-2'], 'public-url-2')
        self.assertEqual(3, len(calls))
        self.assertEqual(('server_started', ['backing-url-2'],
                          'public-url-2', thread), calls[2])

        # Calling on_server_stopped() will call appropriate hooks and
        # eventually shut down the mainloop thread.
        server_mainloop.on_server_stopped(['backing-url-1'], 'public-url-1')
        self.assertEqual(4, len(calls))
        self.assertEqual(('server_stopped', ['backing-url-1'],
                          'public-url-1', thread), calls[3])
        server_mainloop.on_server_stopped(['backing-url-2'], 'public-url-2')
        self.assertEqual(6, len(calls))
        self.assertEqual(('server_stopped', ['backing-url-2'],
                          'public-url-2', thread), calls[4])
        self.assertEqual(('thread_stopped', mainloop, thread), calls[5])
        self.assertEqual(0, mainloop.start_count)
        self.assertEqual(None, mainloop.thread)

    def test_install_hooks(self):
        """install_hooks() should install hooks."""
        server_mainloop.install_hooks()
        self.assertTrue(server_mainloop.on_server_started in
                        SmartTCPServer.hooks['server_started'])
        self.assertTrue(server_mainloop.on_server_stopped in
                        SmartTCPServer.hooks['server_stopped'])

    def test_hooks_are_installed(self):
        """The hooks are installed by default."""
        self.assertTrue(server_mainloop.on_server_started in
            self._preserved_hooks[SmartTCPServer]['server_started'])
        self.assertTrue(server_mainloop.on_server_stopped in
            self._preserved_hooks[SmartTCPServer]['server_stopped'])
